"""
Make formatted files for Gephi Network Visualization

@organization: Covert Lab, Department of Bioengineering, Stanford University
"""

from __future__ import absolute_import, division

import os
import csv


def get_loose_nodes(stoichiometry):
	# get all needed exchanges
	nodes, edges = make_network(stoichiometry)

	sources = set()
	targets = set()
	for edge in edges:
		source = edge[0]
		target = edge[1]
		sources.add(source)
		targets.add(target)

	loose_nodes = sources.symmetric_difference(targets)

	return loose_nodes

def get_reactions(stoichiometry, molecules):
	'''
	for each entry in molecules (list), return all the reactions with the molecules coefficient
	'''
	reactions = {}
	for reaction_id, stoich in stoichiometry.items():
		mol_coeffs = {mol_id: coeff
				for mol_id, coeff in stoich.items()
				if mol_id in molecules}
		reactions[reaction_id] = mol_coeffs

	return reactions

def make_network(stoichiometry, info={}):
	'''
	Makes a gephi network
	info can contain node_sizes, node_types
	info = {
		'node_sizes': node_sizes (dict),
		'node_types': node_types (dict)
	}
	'''

	node_types = info.get('node_types', {})
	node_sizes = info.get('node_sizes', {})
	reaction_fluxes = info.get('reaction_fluxes', {})

	nodes = {}
	edges = []
	for reaction_id, stoich in stoichiometry.items():
		flux = reaction_fluxes.get(reaction_id, 1)
		# add reaction to node list
		n_type = node_types.get(reaction_id, 'reaction')
		n_size = node_sizes.get(reaction_id, 1)
		nodes[reaction_id] = {
			'label': reaction_id,
			'type': n_type,
			'size': n_size}

		# add molecules to node list, and connections to edge list
		for molecule_id, coeff in stoich.items():
			n_type = node_types.get(molecule_id, 'molecule')
			n_size = node_sizes.get(molecule_id, 1)
			nodes[molecule_id] = {
				'label': molecule_id,
				'type': n_type,
				'size': n_size}

			## add edge between reaction and molecule
			# a reactant
			if coeff < 0:
				edge = [molecule_id, reaction_id, flux]
			# a product
			elif coeff > 0:
				edge = [reaction_id, molecule_id, flux]
			else:
				print(reaction_id + ', ' + molecule_id + ': coeff = 0')
				break
			edges.append(edge)

	return nodes, edges

def collapse_network(nodes, edges, remove_nodes_list):
	''' remove_nodes (list) -- nodes to be removed '''
	new_nodes = nodes.copy()

	remove_edges = []
	add_edges = []
	for node_id in remove_nodes_list:
		# remove node from from new_nodes
		new_nodes.pop(node_id, None)

		# remove all edges from and to this node_id
		from_edges = [[from_node, to_node] for [from_node, to_node] in edges if to_node is node_id]
		to_edges = [[from_node, to_node] for [from_node, to_node] in edges if from_node is node_id]
		remove_edges.extend(from_edges)
		remove_edges.extend(to_edges)

		# connect the severed nodes
		for [from_node, c_node1] in from_edges:
			make_edges = [[from_node, to_node] for [c_node2, to_node] in to_edges]
			add_edges.extend(make_edges)

	# make new edges
	new_edges = [edge for edge in edges if edge not in remove_edges]
	new_edges.extend(add_edges)

	# remove redundant new edges
	new_edges2 = []
	for edge in new_edges:
		if edge not in new_edges2:
			new_edges2.append(edge)

	return new_nodes, new_edges2

def save_network(nodes, edges, plotOutDir='out/network'):
	'''
	Save nodes and edges

	Requires:
		- nodes (dict) with {node_id: {'label' (str), 'type' (str), size (float)}}
		- edges (list) with [[node_id1, node_id2] ...]
	'''

	out_dir = os.path.join(plotOutDir)
	if not os.path.exists(out_dir):
		os.makedirs(out_dir)
	nodes_out = os.path.join(out_dir, 'nodes.csv')
	edges_out = os.path.join(out_dir, 'edges.csv')

	## Save network to csv
	# nodes list
	with open(nodes_out, 'w') as csvfile:
		writer = csv.writer(csvfile, delimiter=',')

		# write header
		writer.writerow(['Id', 'Label', 'Type', 'Size'])

		for node, specs in nodes.items():
			label = specs['label']
			type = specs['type']
			size = specs['size']

			row = [node, label, type, size]
			writer.writerow(row)

	# edges list
	with open(edges_out, 'w') as csvfile:
		writer = csv.writer(csvfile, delimiter=',')

		# write header
		writer.writerow(['Source', 'Target', 'Weight'])

		for edge in edges:
			source = edge[0]
			target = edge[1]
			weight = edge[2]

			row = [source, target, weight]
			writer.writerow(row)



from __future__ import absolute_import, division, print_function

import os

from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    plot_simulation_output,
    PROCESS_OUT_DIR,
)


# a global NAME is used for the output directory and for the process name
NAME = 'template'


class Template(Process):
    '''
    This mock process provides a basic template that can be used for a new process
    '''

    # give the process a name, so that it can register in the process_repository
    name = NAME

    # declare default parameters as class variables
    defaults = {
        'parameter_1': {}
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        # get the parameters out of initial_parameters if available, or use defaults
        parameter_1 = self.or_default(
            initial_parameters, 'parameter_1')

        parameters = {'parameter_1': parameter_1}
        super(Template, self).__init__(parameters)

    def ports_schema(self):
        '''
        ports_schema returns a dictionary that declares how each state will behave.
        Each key can be assigned settings for the schema_keys declared in Store:
            '_default'
            '_updater'
            '_divider'
            '_value'
            '_properties'
            '_emit'
            '_serializer'
        '''

        return {
            'internal': {
                'A': {
                    '_default': 1.0,
                    '_updater': 'accumulate',
                    '_emit': True,
                }
            },
            'external': {
                'B': {
                    '_default': 1.0,
                    '_updater': 'accumulate',
                    '_emit': True,
                }
            },
        }

    def derivers(self):
        '''
        declare which derivers are needed for this process
        '''
        return {}

    def next_update(self, timestep, states):
        # get the states
        A_state = states['internal']['A']
        B_state = states['external']['B']

        # calculate output for the states
        A_update = 0.5 * B_state
        B_update = -0.5 * B_state

        # return an update to the states
        return {
            'internal': {'A': A_update},
            'external': {'B': B_update}}


# functions to configure and run the process
def run_template_process(out_dir='out'):
    # initialize the process by passing initial_parameters
    initial_parameters = {}
    template_process = Template(initial_parameters)

    # run the simulation
    sim_settings = {'total_time': 10}
    output = simulate_process_in_experiment(template_process, sim_settings)

    # plot the simulation output
    plot_settings = {}
    plot_simulation_output(output, plot_settings, out_dir)


# run module is run as the main program with python vivarium/process/template_process.py
if __name__ == '__main__':
    # make an output directory to save plots
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    run_template_process(out_dir)



from __future__ import absolute_import, division, print_function

import copy

import numpy as np
from scipy import constants

from vivarium.library.units import units
from vivarium.library.dict_utils import deep_merge
from vivarium.core.process import Deriver

AVOGADRO = constants.N_A * 1 / units.mol


class NonSpatialEnvironment(Deriver):
    '''A non-spatial environment with volume'''

    name = 'nonspatial_environment'
    defaults = {
        'volume': 1e-12 * units.L,
    }

    def __init__(self, parameters=None):
        super(NonSpatialEnvironment, self).__init__(parameters)
        volume = parameters.get('volume', self.defaults['volume'])
        self.mmol_to_counts = (AVOGADRO.to('1/mmol') * volume).to('L/mmol')


    def ports_schema(self):
        bin_x = 1 * units.um
        bin_y = 1 * units.um
        depth = self.parameters['volume'] / bin_x / bin_y
        n_bin_x = 1
        n_bin_y = 1
        return {
            'external': {
                '*': {
                    '_value': 0,
                },
            },
            'fields': {
                '*': {
                    '_value': np.ones((1, 1)),
                },
            },
            'dimensions': {
                'depth': {
                    '_value': depth.to(units.um).magnitude,
                },
                'n_bins': {
                    '_value': [n_bin_x, n_bin_y],
                },
                'bounds': {
                    '_value': [
                        n_bin_x * bin_x.to(units.um).magnitude,
                        n_bin_y * bin_y.to(units.um).magnitude,
                    ],
                },
            },
            'global': {
                'location': {
                    '_value': [0.5, 0.5],
                },
                'volume': {
                    '_value': self.parameters['volume'],
                }
            },
        }

    def next_update(self, timestep, states):
        fields = states['fields']

        update = {
            'external': {
                mol_id: {
                    '_updater': 'set',
                    '_value': field[0][0],
                }
                for mol_id, field in fields.items()
            },
        }

        return update



import os
import glob
import setuptools
from distutils.core import setup

with open("README.md", 'r') as readme:
    long_description = readme.read()

setup(
    name='vivarium-core',
    version='0.0.12',
    packages=[
        'vivarium',
        'vivarium.core',
        'vivarium.processes',
        'vivarium.library'
    ],
    author='Eran Agmon, Ryan Spangler',
    author_email='eagmon@stanford.edu, ryan.spangler@gmail.com',
    url='https://github.com/vivarium-collective/vivarium-core',
    license='MIT',
    entry_points={
        'console_scripts': []},
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data={},
    include_package_data=True,
    install_requires=[
        'confluent-kafka',
        'matplotlib',
        'networkx',
        'numpy',
        'Pint',
        'pymongo',
        'pytest',
        'scipy'])



from __future__ import absolute_import, division, print_function

import uuid
import logging as log

from vivarium.core.process import Deriver


# functions for generating daughter ids
def daughter_uuid(mother_id):
    return [
        str(uuid.uuid1()),
        str(uuid.uuid1())]

def daughter_phylogeny_id(mother_id):
    return [
        str(mother_id) + '0',
        str(mother_id) + '1']


def divider_set_false(state):
    return [False, False]


class MetaDivision(Deriver):

    name = 'meta_division'
    defaults = {
        'initial_state': {},
        'daughter_path': ('agents',),
        'daughter_ids_function': daughter_phylogeny_id}

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        self.division = 0

        # must provide a compartment to generate new daughters
        self.agent_id = initial_parameters['agent_id']
        self.compartment = initial_parameters['compartment']
        self.daughter_ids_function = self.or_default(
            initial_parameters, 'daughter_ids_function')
        self.daughter_path = self.or_default(
            initial_parameters, 'daughter_path')

        super(MetaDivision, self).__init__(initial_parameters)

    def ports_schema(self):
        return {
            'global': {
                'divide': {
                    '_default': False,
                    '_updater': 'set',
                    '_divider': divider_set_false,
                }},
            'agents': {
                '*': {}}}

    def next_update(self, timestep, states):
        divide = states['global']['divide']

        if divide:
            daughter_ids = self.daughter_ids_function(self.agent_id)
            daughter_updates = []
            
            for daughter_id in daughter_ids:
                compartment = self.compartment.generate({
                    'agent_id': daughter_id})
                daughter_updates.append({
                    'daughter': daughter_id,
                    'path': (daughter_id,) + self.daughter_path,
                    'processes': compartment['processes'],
                    'topology': compartment['topology'],
                    'initial_state': {}})

            log.info(
                'DIVIDE! \n--> MOTHER: {} \n--> DAUGHTERS: {}'.format(
                    self.agent_id, daughter_ids))

            # initial state will be provided by division in the tree
            return {
                'agents': {
                    '_divide': {
                        'mother': self.agent_id,
                        'daughters': daughter_updates}}}
        else:
             return {}   
